"""
    tot_footprint_avg_height(conn::MySQL.Connection, selection::Dict) -> DataFrame

Retrieve the total footprint and average height of buildings for specified geographic regions from a MySQL database.

This function executes two SQL queries to calculate the total footprint area and the average height of buildings within specified geographic regions. The geographic regions are defined by the shapes and IDs provided in the `selection` dictionary.

# Arguments
- `conn::MySQL.Connection`: An active connection to a MySQL database.
- `selection::Dict`: A dictionary containing `geo_id` and `shape_obj`, representing geographic identifiers and their corresponding shapes.

# Returns
- `DataFrame`: A DataFrame with columns for the geographic ID, total footprint area, average height of buildings.

# Example
> conn = MySQL.connect("localhost", "username", "password", db = "dbname")
> selection = Dict("geo_id" => ["ID1", "ID2"], "shape_obj" => ["SHAPE1", "SHAPE2"])
> df = tot_footprint_avg_height(conn, selection)

This example queries the total footprint information and average height for specified regions and returns the results in a DataFrame.
"""
function footprint_avg_height(conn_dict::Dict, selection::Dict)
    # getting valuable info from dictionary
    geo_id = selection["geo_id"]
    shape_obj = selection["shape_obj"]
    # object to be returned
    dftemp = DataFrame()
    # iterate over geo objects
    for (id, shape) in zip(geo_id, shape_obj)
        # prepare query
        # given the CRS of the shape, the ST_Area will return squared degrees, hence here the max_footptin and avg_footprint are in millions of squared degrees.
        query = "SELECT IF(ST_Area(ST_GeomFromText('"*shape*"')) = 0, 0, SUM(ST_Area(SHAPE))/ST_Area(ST_GeomFromText('"*shape*"'))) as perc_footprint, MAX(ST_Area(SHAPE))*1000000 as max_footprint, AVG(ST_Area(SHAPE))*1000000 as avg_footprint FROM footprints WHERE ST_Intersects(ST_GeomFromText('"*shape*"'), SHAPE) AND part_id = LEFT('"*id*"', 2);"
        query2 = "SELECT CASE WHEN SUM(CASE WHEN JSON_EXTRACT(properties, '\$.height') <> -1 THEN ST_Area(SHAPE) ELSE 0 END) = 0 THEN NULL ELSE SUM(CASE WHEN JSON_EXTRACT(properties, '\$.height') <> -1 THEN JSON_EXTRACT(properties, '\$.height') * ST_Area(SHAPE) ELSE 0 END) / SUM(CASE WHEN JSON_EXTRACT(properties, '\$.height') <> -1 THEN ST_Area(SHAPE) ELSE 0 END) END as avg_height FROM footprints WHERE ST_Intersects(ST_GeomFromText('"*shape*"'), SHAPE) AND part_id = LEFT('"*id*"', 2);"
        # create connection
        conn = establish_connection(conn_dict["host"], conn_dict["user"], conn_dict["password"], conn_dict["database"]; port=conn_dict["port"])
        # execute
        result = DataFrame()
        try
            # Execute session variable adjustments
            result = query_connection(conn, query)
        catch ex
            @error "Failed to set session variables for query $query with expection " exception=(ex, catch_backtrace())
        end
        # execute
        result2 = DataFrame()
        try
            # Execute session variable adjustments
            result2 = query_connection(conn, query2)
        catch ex
            @error "Failed to set session variables for query $query2 with expection " exception=(ex, catch_backtrace())
        end
        # close connection
        close_connection(conn)
        # execute
        result = hcat(result, result2)
        # Check if the result is empty
        if !isempty(result)
            # add geo_id col
            result[!, :geo_id] = fill(id, nrow(result))
            # copy results to be returned
            dftemp = vcat(dftemp, result)
        else
            println("No data returned from the query:")
            println(query)
        end
    end 
    #dftemp[!, :region] = fill("EU", nrow(dftemp))
    return dftemp
end

"""
    get_bldgs_footprint_height(table::String, selection::Dict; host::String="127.0.0.1") -> DataFrame

Retrieve building footprint and height data based on the specified table and selection criteria.

This function selects data from a database based on the provided `table` argument and retrieves building footprint and height data matching the given `selection` dictionary. It queries data using predefined SQL logic tailored for buildings' footprints and average heights.

# Arguments
- `table::String`: Specifies the database table or the predefined query identifier to be used for data retrieval (e.g., 'footprint_avg_height').
- `selection::Dict`: A dictionary containing selection criteria to filter the data (e.g., geographic IDs or specific shapes).
- `host::String` (optional): Specifies the hostname of the MySQL server, with a default value of '127.0.0.1'.

# Returns
- `DataFrame`: A DataFrame containing the queried building data. Columns include geographic ID (`geo_id`), percentage footprint (`perc_footprint`), maximum footprint (`max_footprint`), average footprint (`avg_footprint`), and average height (`avg_height`) of buildings within the selected area.

# Example
> selection = Dict("geo_id" => ["ID1"], "shape_obj" => ["SHAPE1"])
> df = get_bldgs_footprint_height("footprint_avg_height", selection)

In this example, the function retrieves building footprint and height data for the specified selection criteria and returns it in a DataFrame.
"""
function get_bldgs_footprint_height(table::String, selection::Dict, system::OrchestratorRegistry.ServiceRegistry)
    service = get_services_by_name(system, "FACT_bldgs")
    conn_dict = Dict(
        "host" => service.host,
        "port" => service.port,
        "database" => service.name,
        "user" => "root",
        "password" => "devops"
    )

    table_dict = Dict(
        "footprint_avg_height" => footprint_avg_height
    )

    df_template = DataFrame(
        geo_id = String[],  # Column for Geo IDs
        perc_footprint = Float64[], # column for percentage of area covered by total footprint of buildings
        max_footprint = Int64[], # column for max footprint of buildings, in millions of squared degrees
        avg_footprint = Float64[], # column for average footprint of buildings, in millions of squared degrees
        avg_height = Float64[] # column to average height of buildings
    )

    # Check if the table is supported
    if !haskey(table_dict, table)
        error("Unsupported table: $table. Please check the available tables: $(keys(table_dict)).")
    end

    # data retrieval
    df = get_selected_objects_from_table(table, df_template, table_dict, selection, conn_dict) 

    return df
end

