"""
    geology_full(conn::MySQL.Connection, selection::Dict) -> DataFrame

Retrieve geology data for specified geographic areas and attributes.

This function retrieves geology data for specified geographic areas based on the provided selection criteria. It communicates with a MySQL database connection to execute SQL queries and returns the results as a DataFrame.

# Arguments
- `conn::MySQL.Connection`: The MySQL database connection to use for querying geology data.
- `selection::Dict`: A dictionary specifying the selection criteria including `geo_id`, `shape_obj`, `region`, and `area`.

# Returns
- `DataFrame`: A DataFrame containing columns for `geo_id` (geographic ID), `region` (region identifier), and various attributes representing the percentage of geological eras and rock types within the specified geographic areas.

# Example
> conn = MySQL.Connection(user="user", password="password", host="localhost", db="geology_data")
> selection = Dict("geo_id" => ["geo1", "geo2"], "shape_obj" => ["WKT1", "WKT2"], "region" => "US", "area" => [1000, 2000])
> df = geology_full(conn, selection)

This example retrieves geology data for two geographic areas in the "US" region with specified shape data and areas.

"""
function geology_full(conn_dict::Dict, selection::Dict)
    # getting valuable info from dictionary
    geo_id = selection["geo_id"]
    shape_obj = selection["shape_obj"]
    #region = selection["region"]
    area = selection["area"]
    # object to be returned
    dftemp = DataFrame()
    # iterate over geo objects
    for (id, shape, ar) in zip(geo_id, shape_obj, area)
        # prepare query
        query = """
        SELECT SUM(CASE WHEN era = 'Cenozoic' THEN ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape'))) ELSE 0 END) / $ar as cenozoic, \
        SUM(CASE WHEN era = 'Mesozoic' THEN ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape'))) ELSE 0 END) / $ar as mesozoic, \
        SUM(CASE WHEN era = 'Paleozoic-Mesozoic' THEN ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape'))) ELSE 0 END) / $ar as paleoMeso, \
        SUM(CASE WHEN era = 'Paleozoic' THEN ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape'))) ELSE 0 END) / $ar as paleozoic, \
        SUM(CASE WHEN era = 'Mesozoic-Cenozoic' THEN ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape'))) ELSE 0 END) / $ar as mesoCeno, \
        SUM(CASE WHEN era = 'Precambrian' THEN ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape'))) ELSE 0 END) / $ar as precambrian, \
        SUM(CASE WHEN era = 'Neoproterozoic' THEN ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape'))) ELSE 0 END) / $ar as neoproterozoic, \
        SUM(CASE WHEN era = 'Archean-Paleoproterozoic' THEN ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape'))) ELSE 0 END) / $ar as archPaleoprotero, \
        SUM(CASE WHEN era = 'Mesoproterozoic' THEN ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape'))) ELSE 0 END) / $ar as mesoproterozoic, \
        SUM(CASE WHEN era = 'Paleoproterozoic' THEN ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape'))) ELSE 0 END) / $ar as paleoproterozoic, \
        SUM(CASE WHEN era = 'Archean' THEN ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape'))) ELSE 0 END) / $ar as archean, \
        SUM(CASE WHEN era = 'Paleoproterozoic-Mesoproterozoic' THEN ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape'))) ELSE 0 END) / $ar as paleoproteroMesoprotero, \
        SUM(CASE WHEN era = 'Mesoproterozoic-Neoproterozoic' THEN ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape'))) ELSE 0 END) / $ar as mesoproteroNeoprotero, \
        SUM(CASE WHEN era = 'Precambrian-Phanerozoic' THEN ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape'))) ELSE 0 END) / $ar as precamPhanero, \
        SUM(CASE WHEN era = 'Phanerozoic' THEN ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape'))) ELSE 0 END) / $ar as phanerozoic, \
        SUM(CASE WHEN era = 'Proterozoic' THEN ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape'))) ELSE 0 END) / $ar as proterozoic, \
        SUM(CASE WHEN era = 'Eoarchean' THEN ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape'))) ELSE 0 END) / $ar as eoarchean, \
        SUM(CASE WHEN rxtyp = 'volcanic rocks' THEN ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape'))) ELSE 0 END) / $ar as volcanic, \
        SUM(CASE WHEN rxtyp = 'sedimentary rocks' THEN ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape'))) ELSE 0 END) / $ar as sedimentary, \
        SUM(CASE WHEN rxtyp = 'intrusive rocks' THEN ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape'))) ELSE 0 END) / $ar as intrusive, \
        SUM(CASE WHEN rxtyp = 'crystalline metamorphic rocks' THEN ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape'))) ELSE 0 END) / $ar as crystalline, \
        SUM(CASE WHEN rxtyp = 'sedimentary and volcanic rocks' THEN ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape'))) ELSE 0 END) / $ar as sediVolc, \
        SUM(CASE WHEN rxtyp = 'volcanic and intrusive rocks' THEN ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape'))) ELSE 0 END) / $ar as volcIntr, \
        SUM(CASE WHEN rxtyp = 'ice' THEN ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape'))) ELSE 0 END) / $ar as ice, \
        SUM(CASE WHEN rxtyp = 'tectonic rocks' THEN ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape'))) ELSE 0 END) / $ar as tectonic, \
        SUM(CASE WHEN rxtyp = 'unclassified rocks' THEN ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape'))) ELSE 0 END) / $ar as unclassified \
        FROM geology WHERE ST_Intersects(ST_GeomFromText('$shape'), SHAPE);
        """
        # create connection
        conn = establish_connection(conn_dict["host"], conn_dict["user"], conn_dict["password"], conn_dict["database"]; port=conn_dict["port"])
        # execute
        result = DataFrame()
        try
            # Execute session variable adjustments
            result = query_connection(conn, query)
        catch ex
            @error "Failed to set session variables for query $query with expection " exception=(ex, catch_backtrace())
        end
        # close connection
        close_connection(conn)
        # Check if the result is empty
        if !isempty(result)
            # add geo_id col
            result[!, :geo_id] = fill(id, nrow(result))
            #result[!, :region] = fill(region, nrow(result))
            # copy results to be returned
            dftemp = vcat(dftemp, result)
        else
            println("No data returned from the query:")
            println(query)
        end
    end 
    return dftemp
end

"""
    minerals_full(conn::MySQL.Connection, selection::Dict) -> DataFrame

Retrieve mineral data for specified geographic areas and attributes.

This function retrieves mineral data for specified geographic areas based on the provided selection criteria. It communicates with a MySQL database connection to execute SQL queries and returns the results as a DataFrame.

# Arguments
- `conn::MySQL.Connection`: The MySQL database connection to use for querying mineral data.
- `selection::Dict`: A dictionary specifying the selection criteria including `geo_id`, `shape_obj`, `region`, and `area`.

# Returns
- `DataFrame`: A DataFrame containing columns for `geo_id` (geographic ID), `region` (region identifier), and various attributes representing the percentage area of different mineral types within the specified geographic areas.

# Example
> conn = MySQL.Connection(user="user", password="password", host="localhost", db="mineral_data")
> selection = Dict("geo_id" => ["geo1", "geo2"], "shape_obj" => ["WKT1", "WKT2"], "region" => "EU", "area" => [1500, 2500])
> df = minerals_full(conn, selection)

This example retrieves mineral data for two geographic areas in the "EU" region with specified shape data and areas.
"""
function minerals_full(conn_dict::Dict, selection::Dict)
    # getting valuable info from dictionary
    geo_id = selection["geo_id"]
    shape_obj = selection["shape_obj"]
    #region = selection["region"]
    area = selection["area"]
    # object to be returned
    dftemp = DataFrame()
    # iterate over geo objects
    for (id, shape, ar) in zip(geo_id, shape_obj, area)
        # create connection
        conn = establish_connection(conn_dict["host"], conn_dict["user"], conn_dict["password"], conn_dict["database"]; port=conn_dict["port"])
        tables = ["carbonatite" , "global_selected" , "laterite" , "nicrpge" , "phosphate" , "podchrome" , "porcu" , "potash" , "ree" , "sedcu" , "sedexmvt" , "sedznpb" , "vms"]
        result = DataFrame()
        for table in tables
            # prepare query
            query = "SELECT SUM(ST_Area(ST_Intersection(SHAPE, ST_GeomFromText('$shape')))) / $ar AS $table FROM $table WHERE ST_Intersects(SHAPE, ST_GeomFromText('$shape'));"
            # execute
            result2 = DataFrame()
            try
                # Execute session variable adjustments
                result2 = query_connection(conn, query)
            catch ex
                @error "Failed to set session variables for query $query with expection " exception=(ex, catch_backtrace())
            end
            result = hcat(result, result2)
        end
        # close connection
        close_connection(conn)
        # Check if the result is empty
        if !isempty(result)
            # add geo_id col
            result[!, :geo_id] = fill(id, nrow(result))
            #result[!, :region] = fill(region, nrow(result))
            # copy results to be returned
            dftemp = vcat(dftemp, result)
        else
            println("No data returned from the query:")
            println(query)
        end
    end 
    #dftemp[!, :region] = fill("EU", nrow(dftemp))
    return dftemp
end

"""
    get_geology(table::String, selection::Dict; host::String="127.0.0.1") -> DataFrame

Retrieve geology data for specified geographic areas and attributes using a service query.

This function allows for the retrieval of geology data from a specific table using a service query and selection criteria. It communicates with the MySQL database and returns the results as a DataFrame.

# Arguments
- `table::String`: The name of the table to query for geology data.
- `selection::Dict`: A dictionary specifying the selection criteria for the query, including `geo_id`, `shape_obj`, `region`, and `area`.
- `host::String` (optional): The host address of the MySQL database. Default is "127.0.0.1".

# Returns
- `DataFrame`: A DataFrame containing columns for `geo_id` (geographic ID), `region` (region identifier), and various attributes representing the percentage of geological eras and rock types within the specified geographic areas.

# Example
> table = "geology_data"
> selection = Dict("geo_id" => ["geo1", "geo2"], "shape_obj" => ["WKT1", "WKT2"], "region" => "US", "area" => [1000, 2000])
> df = get_geology(table, selection, host="localhost")

This example retrieves geology data using the specified table and selection criteria.
"""
function get_geology(table::String, selection::Dict, system::OrchestratorRegistry.ServiceRegistry)
    service = get_services_by_name(system, "FACT_mine")
    conn_dict = Dict(
        "host" => service.host,
        "port" => service.port,
        "database" => service.name,
        "user" => "root",
        "password" => "devops"
    )
    table_dict = Dict(
        "geology_full" => geology_full
    )

    df_template = DataFrame(
        #region = String[], # column to store if EU or US
        geo_id = String[],  # Column for Geo IDs
        cenozoic = Float64[], # column for percentage of geology with era = Cenozoic
        mesozoic = Float64[], # column for percentage of geology with era = Mesozoic
        paleoMeso = Float64[], # column for percentage of geology with era = Paleozoic-Mesozoic
        paleozoic = Float64[], # column for percentage of geology with era = Paleozoic
        mesoCeno = Float64[], # column for percentage of geology with era = Mesozoic-Cenozoic
        precambrian = Float64[], # column for percentage of geology with era = Precambrian
        neoproterozoic = Float64[], # column for percentage of geology with era = Neoproterozoic
        archPaleoprotero = Float64[], # column for percentage of geology with era = Archean-Paleoproterozoic
        mesoproterozoic = Float64[], # column for percentage of geology with era = Mesoproterozoic
        paleoproterozoic = Float64[], # column for percentage of geology with era = Paleoproterozoic
        archean = Float64[], # column for percentage of geology with era = Archean
        paleoproteroMesoprotero = Float64[], # column for percentage of geology with era = Paleoproterozoic-Mesoproterozoic
        mesoproteroNeoprotero = Float64[], # column for percentage of geology with era = Mesoproterozoic-Neoproterozoic
        precamPhanero = Float64[], # column for percentage of geology with era = Precambrian-Phanerozoic
        phanerozoic = Float64[], # column for percentage of geology with era = Phanerozoic
        proterozoic = Float64[], # column for percentage of geology with era = Proterozoic
        eoarchean = Float64[], # column for percentage of geology with era = Eoarchean
        volcanic = Float64[], # column for percentage of volcanic rocks
        sedimentary = Float64[], # column for percentage of sedimentary rocks
        intrusive = Float64[], # column for percentage of intrusive rocks
        crystalline = Float64[], # column for percentage of crystalline metamorphic rocks
        sediVolc = Float64[], # column for percentage of sedimentary and volcanic rocks
        volcIntr = Float64[], # column for percentage of volcanic and intrusive rocks
        ice = Float64[], # column for percentage of ice
        tectonic = Float64[], # column for percentage of tectonic rocks
        unclassified = Float64[] # column for percentage of unclassified rocks
    )

    # Check if the table is supported
    if !haskey(table_dict, table)
        error("Unsupported table: $table. Please check the available tables: $(keys(table_dict)).")
    end

    # data retrieval
    df = get_selected_objects_from_table(table, df_template, table_dict, selection, conn_dict) 

    return df
end

"""
    get_minerals(table::String, selection::Dict; host::String="127.0.0.1") -> DataFrame

Retrieve mineral data for specified geographic areas and attributes using a service query.

This function allows for the retrieval of mineral data from a specific table using a service query and selection criteria. It communicates with the MySQL database and returns the results as a DataFrame.

# Arguments
- `table::String`: The name of the table to query for mineral data.
- `selection::Dict`: A dictionary specifying the selection criteria for the query, including `geo_id`, `shape_obj`, `region`, and `area`.
- `host::String` (optional): The host address of the MySQL database. Default is "127.0.0.1".

# Returns
- `DataFrame`: A DataFrame containing columns for `geo_id` (geographic ID), `region` (region identifier), and various attributes representing the percentage area of different mineral types within the specified geographic areas.

# Example
> table = "mineral_data"
> selection = Dict("geo_id" => ["geo1", "geo2"], "shape_obj" => ["WKT1", "WKT2"], "region" => "EU", "area" => [1500, 2500])
> df = get_minerals(table, selection, host="localhost")

This example retrieves mineral data using the specified table and selection criteria.
"""
function get_minerals(table::String, selection::Dict, system::OrchestratorRegistry.ServiceRegistry)
    service = get_services_by_name(system, "FACT_mine")
    conn_dict = Dict(
        "host" => service.host,
        "port" => service.port,
        "database" => service.name,
        "user" => "root",
        "password" => "devops"
    )

    table_dict = Dict(
        "minerals_full" => minerals_full
    )

    df_template = DataFrame(
        #region = String[], # column to store if EU or US
        geo_id = String[],  # Column for Geo IDs
        carbonatite = Float64[], # column for percentage area of ...
        global_selected = Float64[], # column for percentage area of ...
        laterite = Float64[], # column for percentage area of ...
        nicrpge = Float64[], # column for percentage area of ...
        phosphate = Float64[], # column for percentage area of ...
        podchrome = Float64[], # column for percentage area of ...
        porcu = Float64[], # column for percentage area of ...
        potash = Float64[], # column for percentage area of ...
        ree = Float64[], # column for percentage area of ...
        sedcu  = Float64[], # column for percentage area of ...
        sedexmvt  = Float64[], # column for percentage area of ...
        sedznpb = Float64[], # column for percentage area of ...
        vms = Float64[] # column for percentage area of ...
    )

    # Check if the table is supported
    if !haskey(table_dict, table)
        error("Unsupported table: $table. Please check the available tables: $(keys(table_dict)).")
    end

    # data retrieval
    df = get_selected_objects_from_table(table, df_template, table_dict, selection, conn_dict) 

    return df
end
