"""
    travel_time_in_minutes(ws_host_port::String, selection::Dict) -> DataFrame

Calculate travel times in minutes between pairs of geographic locations.

This function calculates travel times in minutes between pairs of geographic locations using the Mapbox Directions API. It takes a Mapbox Directions API endpoint as `ws_host_port` and a selection of geographic locations specified as `geo_id` and `shape_obj` in a dictionary. The travel times are computed for all possible pairs of locations and stored in a DataFrame.

# Arguments
- `ws_host_port::String`: The URL of the Mapbox Directions API endpoint to use for calculating travel times.
- `selection::Dict`: A dictionary specifying the selection criteria including `geo_id` (geographic IDs) and `shape_obj` (geometric shapes) for the locations.

# Returns
- `DataFrame`: A DataFrame containing columns for `o_geo_id` (origin geographic ID), `d_geo_id` (destination geographic ID), and `travel_time` (travel time in minutes) representing travel times between pairs of geographic locations.

# Example
> ws_host_port = "https://api.mapbox.com/directions/v5/mapbox/driving"
> selection = Dict(
>     "geo_id" => ["Location1", "Location2", "Location3"],
>     "shape_obj" => ["POINT(1 2)", "POINT(3 4)", "POINT(5 6)"]
> )
> df = travel_time_in_minutes(ws_host_port, selection)

This example calculates travel times in minutes between three geographic locations using the Mapbox Directions API.

"""
function travel_time_in_minutes(ws_host_port::String, selection::Dict)
    # getting valuable info from dictionary
    geo_id = selection["geo_id"]
    shape_obj = selection["shape_obj"]

    # Arrays to store column data
    o_geo_ids = String[]
    d_geo_ids = String[]
    travel_times = Float64[]
    # iterate over geo objects
    centroids = []
    for shape in shape_obj
        # compute centroids
        centroid_geom = centroid(readgeom(shape))
        push!(centroids, centroid_geom)
    end
    for i in 1:lastindex(centroids)
        for j in i+1:lastindex(centroids)
            start_longitude = LibGEOS.getcoord(centroids[i], 1)
            start_latitude = LibGEOS.getcoord(centroids[i], 2)
            end_longitude = LibGEOS.getcoord(centroids[j], 1)
            end_latitude = LibGEOS.getcoord(centroids[j], 2)

            # Construct the request URL
            url = "$ws_host_port/route/v1/driving/$(start_longitude),$(start_latitude);$(end_longitude),$(end_latitude)?overview=false"

            travel_time = -9999 #sentinel value

            try
                # Make the request
                response = HTTP.get(url)
                # Ensure the response status code is 200 (OK)
                if response.status != 200
                    throw(Exception("Request failed with status code $(response.status)"))
                end
                # Parse the response
                data = JSON.parse(String(response.body))
                # Extract and calculate travel time in minutes
                travel_time_seconds = data["routes"][1]["duration"]
                travel_time = travel_time_seconds / 60
            catch e
                println("Error during HTTP request or response parsing: ", e)
            end

            # Populate the columns
            push!(o_geo_ids, geo_id[i])
            push!(d_geo_ids, geo_id[j])
            push!(travel_times, travel_time)

            # Add the reverse trip
            push!(o_geo_ids, geo_id[j])
            push!(d_geo_ids, geo_id[i])
            push!(travel_times, travel_time)
        end
    end
    # object to be returned
    dftemp = DataFrame(
        o_geo_id = o_geo_ids,
        d_geo_id = d_geo_ids,
        travel_time = travel_times
    )
    return dftemp
end

"""
    get_travel_data_eu(service_query::String, selection::Dict; host::String="127.0.0.1") -> DataFrame

Retrieve travel time data for European Union (EU) geographic locations using a service query.

This function allows for the retrieval of travel time data between pairs of EU geographic locations using a service query. It communicates with a service provided by the specified host and port, and returns the results as a DataFrame.

# Arguments
- `service_query::String`: The name of the service query to use for retrieving travel time data (e.g., "travel_time_in_minutes").
- `selection::Dict`: A dictionary specifying the selection criteria for the service query, including `geo_id` and `shape_obj`.
- `host::String` (optional): The host address of the service. Default is "127.0.0.1".

# Returns
- `DataFrame`: A DataFrame containing columns for `o_geo_id` (origin geographic ID), `d_geo_id` (destination geographic ID), and `travel_time` (travel time in minutes) representing travel time data for pairs of EU geographic locations.

# Example
> service_query = "travel_time_in_minutes"
> selection = Dict(
>     "geo_id" => ["Location1", "Location2", "Location3"],
>     "shape_obj" => ["POINT(1 2)", "POINT(3 4)", "POINT(5 6)"]
> )
> df = get_travel_data_eu(service_query, selection, host="localhost")

This example retrieves travel time data for pairs of EU geographic locations using the "travel_time_in_minutes" service query.

"""
function get_travel_data_eu(service_query::String, selection::Dict, system::OrchestratorRegistry.ServiceRegistry)
    service = get_services_by_name(system, "FACT_road_eu")
    service_dict = Dict(
        "travel_time_in_minutes" => travel_time_in_minutes
    )
    df_template = DataFrame(
        o_geo_id = String[],  # Column for Origin Geo IDs
        d_geo_id = String[],  # Column for Destination Geo IDs
        travel_time = Float64[] # column for travel time in seconds
    )

    # Check if the service is supported
    if !haskey(service_dict, service_query)
        error("Unsupported service: $service_query. Please check the available services: $(keys(service_dict)).")
    end

    # data retrieval
    df = get_results_from_service(service_query, df_template, service_dict, selection; host=service.host, port=service.port)
    return df
end

"""
    get_travel_data_na(service_query::String, selection::Dict; host::String="127.0.0.1") -> DataFrame

Retrieve travel time data for North America (NA) geographic locations using a service query.

This function allows for the retrieval of travel time data between pairs of North America (NA) geographic locations using a service query. It communicates with a service provided by the specified host and port, and returns the results as a DataFrame.

# Arguments
- `service_query::String`: The name of the service query to use for retrieving travel time data (e.g., "travel_time_in_minutes").
- `selection::Dict`: A dictionary specifying the selection criteria for the service query, including `geo_id` and `shape_obj`.
- `host::String` (optional): The host address of the service. Default is "127.0.0.1".

# Returns
- `DataFrame`: A DataFrame containing columns for `o_geo_id` (origin geographic ID), `d_geo_id` (destination geographic ID), and `travel_time` (travel time in minutes) representing travel time data for pairs of NA geographic locations.

# Example
> service_query = "travel_time_in_minutes"
> selection = Dict(
>     "geo_id" => ["Location1", "Location2", "Location3"],
>     "shape_obj" => ["POINT(1 2)", "POINT(3 4)", "POINT(5 6)"]
> )
> df = get_travel_data_na(service_query, selection, host="localhost")

This example retrieves travel time data for pairs of NA geographic locations using the "travel_time_in_minutes" service query.

"""
function get_travel_data_na(service_query::String, selection::Dict, system::OrchestratorRegistry.ServiceRegistry)
    service = get_services_by_name(system, "FACT_road_na")
    service_dict = Dict(
        "travel_time_in_minutes" => travel_time_in_minutes
    )
    df_template = DataFrame(
        o_geo_id = String[],  # Column for Origin Geo IDs
        d_geo_id = String[],  # Column for Destination Geo IDs
        travel_time = Float64[] # column for travel time in seconds
    )

    # Check if the service is supported
    if !haskey(service_dict, service_query)
        error("Unsupported service: $service_query. Please check the available services: $(keys(service_dict)).")
    end

    # data retrieval
    df = get_results_from_service(service_query, df_template, service_dict, selection; host=service.host, port=service.port)
    return df
end
